"""
Tools for working with :py:class:`traffic.lib.Traffic` cycles.
"""
import operator
import itertools
import collections

#: Singleton Tuple of event identifiers produced by :py:class:`traffic.lib.Transit`
#: instances. Useful for referencing event identifiers without importing the
#: primary library.
events = collections.namedtuple("Events", "terminate transfer exhaust")(*range(3))

def unroll(traffic):
	"""
	unroll(traffic)

	:param traffic: The traffic instance whose window should be unrolled.
	:rtype: generator
	:returns: An iterator producing transit-event pairs: (transit, event).

	Given a :py:class:`traffic.lib.Traffic` instance, yield each event that
	occurred associated with the Transit that produced the event.
	"""
	for t in traffic.transfer():
		for e in t.events():
			yield (t, e)

#: Used to hold a snapshot of events and data that were produced by a Transit and their
#: associated information or callback.
#: 
#: Attributes:
#: 
#:  ``transit``
#:   The :py:class:`traffic.lib.Transit` instance.
#: 
#:  ``link``
#:   The :py:attr:`traffic.lib.Transit.link` attribute of the `transit`.
#: 
#:  ``terminated``
#:   Whether or not a terminate event occurred: `None` if not terminated.
#:   Otherwise, the result of the :py:meth:`traffic.lib.Point.failure` method.
#: 
#:  ``transferred``
#:   An integer designating the amount transferred iff the transit sends.
#:   The transfer slice of the resource iff the transit receives.
#: 
#:  ``demand``
#:   The :py:meth:`traffic.lib.Transit.acquire` method if an exhaust event
#:   occurred, otherwise, `None`.
Activity = collections.namedtuple("Activity",
	("transit", "link", "termination", "transferred", "demand",)
)

def activity(transit,
	term_event = events.terminate,
	xfer_event = events.transfer,
	exht_event = events.exhaust,
	Activity = Activity,
):
	"""
	Given a :py:class:`traffic.lib.Transit`, get an :py:class:`traffic.libloop.Activity`
	snapshot of the Transit.
	"""
	termination = None # not terminated
	demand = None
	xfer = None

	for event in transit.events():
		if event == xfer_event:
			slic = transit.transfer()
			if transit.polarity:
				# data for receives
				xfer = transit.resource[slic]
			else:
				# size for sends
				xfer = slic.stop - slic.start
		elif event == exht_event:
			demand = transit.acquire
			# termination and exhaust follow transfers
			# termination cannot follow exhaust
			break
		else:
			termination = transit.failure()
			# termination and exhaust follow transfers
			# exhaust cannot follow termination
			break

	return Activity(transit, transit.link, termination, xfer, demand)

def snapshot(traffic, _chain=itertools.chain.from_iterable):
	"""
	Given a :py:class:`traffic.lib.Traffic` instance, get a snapshot of all
	:py:func:`traffic.libloop.activity`.
	"""
	# list() is invoked here as materialization of the
	# snapshot is needed in cases where the cycle exits. (like libloop.cycle)
	return list(map(activity, traffic.transfer()))

def cycle(traffic):
	"""
	Cycle the traffic instance and return a :py:func:`traffic.libloop.snapshot`
	list of activity.

	.. warning:: If insufficient memory is available for building the snapshot,
	             transfer-loss *will* occur.
	"""
	with traffic:
		return snapshot(traffic)
